// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "role_side_back.h"

#include "strategy.h"

#include "bhv_basic_offensive_kick.h"
#include "bhv_basic_move.h"

#include <rcsc/formation/formation.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>


#include <rcsc/action/body_advance_ball.h>//advance ball to enemeyfield p72//
#include <rcsc/action/body_clear_ball.h>//clear ball p72//
#include <rcsc/action/body_go_to_point.h>//dash to targetopoint p73//
#include <rcsc/action/body_pass.h>//pass p75//


const std::string RoleSideBack::NAME( "SideBack" );

/*-------------------------------------------------------------------*/
/*!

 */
namespace {
rcss::RegHolder role = SoccerRole::creators().autoReg( &RoleSideBack::create,
                                                       RoleSideBack::NAME );
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideBack::execute( rcsc::PlayerAgent * agent )
{
    bool kickable = agent->world().self().isKickable();
    if ( agent->world().existKickableTeammate()
         && agent->world().teammatesFromBall().front()->distFromBall()
         < agent->world().ball().distFromSelf() )
    {
        kickable = false;
    }

    if ( kickable )
    {
        doKick( agent );
    }
    else
    {
        doMove( agent );
    }

}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideBack::doKick( rcsc::PlayerAgent * agent )
{
    switch ( Strategy::get_ball_area( agent->world().ball().pos() ) ) {
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
    case Strategy::BA_DribbleAttack:
    case Strategy::BA_OffMidField:
    case Strategy::BA_Cross:
    case Strategy::BA_ShootChance:
    default:
        Bhv_BasicOffensiveKick().execute( agent );
        break;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideBack::doMove( rcsc::PlayerAgent * agent )
{
////////add
const rcsc::WorldModel & wm = agent->world();
const rcsc::Vector2D ball_pos=wm.ball().pos();
//std::cout<<"ikame"<<std::endl;
if((ball_pos.x < -25.0 && ball_pos.y < 30.0 )||(ball_pos.x<-25.0 && ball_pos.y>-30.0)){
const int my_unum=wm.self().unum();
	if(my_unum==4 || my_unum==5){ //p160//
	double lrside=0.0;
	if(my_unum ==4){
	lrside=2.0;
	}
	if(my_unum ==5){
	lrside=-2.0;
	}
	const rcsc::Vector2D target_point(-52.2,lrside);
	rcsc::Vector2D my_point = wm.self().playerType().inertiaPoint (wm.self().pos(),wm.self().vel(),60);
	
	rcsc::Vector2D rel_pos = target_point - my_point;
	rcsc::AngleDeg turn_angle =rel_pos.th()-wm.self().body(); 
	double target_dist = rel_pos.r();
	double turn_thr = 180.0;
	double dist_thr=(cos((90.0-turn_thr))*(2.0*(asin(1.0))/180))*target_dist;
	//ekichi p162//
	if(dist_thr < target_dist){ //not targetpoint//
		turn_thr = rcsc::AngleDeg::asin_deg(dist_thr / target_dist);
	 }
	turn_thr = std::max(turn_thr , 2.0);
	if(turn_angle.abs() > turn_thr){
		agent->doTurn (turn_angle);
	 }
	
	//dash targetpoint p163-164//
	rel_pos.rotate( - wm.self().body() );//body to X//
	double first_speed = rcsc::calc_first_term_geom_series( rel_pos.x, wm.self().playerType().playerDecay(), 60);
	first_speed = rcsc::min_max( -wm.self().playerType().playerSpeedMax(),first_speed,wm.self().playerType().playerSpeedMax() );
	const rcsc::Vector2D rel_vel =wm.self().vel().rotatedVector(wm.self().body());
	double accel = first_speed - rel_vel.x;
	accel = rcsc::min_max(- rcsc::ServerParam::i().playerAccelMax(), accel,rcsc::ServerParam::i().playerAccelMax() );
	double dash_power = accel / wm.self().dashRate();
	dash_power = rcsc::ServerParam::i().normalizePower( dash_power );
	agent -> doDash ( dash_power );
	
	//stop to targetpoint p164//
	double dash_powerb = -(rel_vel.x / wm.self().dashRate() );
	dash_powerb = rcsc::ServerParam::i().normalizePower( dash_powerb );
	agent -> doDash( dash_powerb );
	

	}
	}

////////end
    rcsc::Vector2D home_pos = Strategy::i().getPosition( agent->world().self().unum() );
    if ( ! home_pos.valid() ) home_pos.assign( 0.0, 0.0 );

    switch ( Strategy::get_ball_area( agent->world() ) ) {
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
    case Strategy::BA_DribbleAttack:
    case Strategy::BA_OffMidField:
    case Strategy::BA_Cross:
    case Strategy::BA_ShootChance:
    default:
        Bhv_BasicMove( home_pos ).execute( agent );
        break;
    }
}

//2009-2010追加
void 
RoleSideBack::doDefenceback( rcsc::PlayerAgent * agent )
{
const rcsc::WorldModel & wm = agent->world();
const rcsc::Vector2D ball_pos=wm.ball().pos();
//std::cout<<"ikame"<<std::endl;
if((ball_pos.x < -25.0 && ball_pos.y < 30.0 )||(ball_pos.x<-25.0 && ball_pos.y>-30.0)){
const int my_unum=wm.self().unum();
	if(my_unum==4 || my_unum==5){ //p160//
	double lrside=0.0;
	if(my_unum ==4){
	lrside=2.0;
	}
	if(my_unum ==5){
	lrside=-2.0;
	}
	const rcsc::Vector2D target_point(-52.2,lrside);
	rcsc::Vector2D my_point = wm.self().playerType().inertiaPoint (wm.self().pos(),wm.self().vel(),60);
	
	rcsc::Vector2D rel_pos = target_point - my_point;
	rcsc::AngleDeg turn_angle =rel_pos.th()-wm.self().body(); 
	double target_dist = rel_pos.r();
	double turn_thr = 180.0;
	double dist_thr=(cos((90.0-turn_thr))*(2.0*(asin(1.0))/180))*target_dist;
	//ekichi p162//
	if(dist_thr < target_dist){ //not targetpoint//
		turn_thr = rcsc::AngleDeg::asin_deg(dist_thr / target_dist);
	 }
	turn_thr = std::max(turn_thr , 2.0);
	if(turn_angle.abs() > turn_thr){
		agent->doTurn (turn_angle);
	 }
	
	//dash targetpoint p163-164//
	rel_pos.rotate( - wm.self().body() );//body to X//
	double first_speed = rcsc::calc_first_term_geom_series( rel_pos.x, wm.self().playerType().playerDecay(), 60);
	first_speed = rcsc::min_max( -wm.self().playerType().playerSpeedMax(),first_speed,wm.self().playerType().playerSpeedMax() );
	const rcsc::Vector2D rel_vel =wm.self().vel().rotatedVector(wm.self().body());
	double accel = first_speed - rel_vel.x;
	accel = rcsc::min_max(- rcsc::ServerParam::i().playerAccelMax(), accel,rcsc::ServerParam::i().playerAccelMax() );
	double dash_power = accel / wm.self().dashRate();
	dash_power = rcsc::ServerParam::i().normalizePower( dash_power );
	agent -> doDash ( dash_power );
	
	//stop to targetpoint p164//
	double dash_powerb = -(rel_vel.x / wm.self().dashRate() );
	dash_powerb = rcsc::ServerParam::i().normalizePower( dash_powerb );
	agent -> doDash( dash_powerb );
	

	}
	}
}

